import{r as n,f as d,m as l,W as i}from"./index.16929924.js";const a=(e="2.2.3",o=()=>{})=>{const t="__bokeh_lib_injection__",s=document.getElementById(t);if(s)window.Bokeh?o():s.addEventListener("load",o);else{const r=document.createElement("script");r.src=`https://cdn.bokeh.org/bokeh/release/bokeh-${e}.min.js`,r.id=t,document.body.appendChild(r),r.addEventListener("load",o)}},c=e=>{const[o,t]=n.exports.useState(!1);return e.bokehJson?.version?o?(window.Bokeh.logger.set_level("error"),n.exports.createElement(h,{...e})):(a(e.bokehJson.version,()=>{t(!0)}),n.exports.createElement(n.exports.Fragment,null)):n.exports.createElement(n.exports.Fragment,null,"-")},h=e=>{const o=n.exports.useRef(null),t=n.exports.useMemo(()=>"_bokeh__"+new Date().getTime(),[]);return n.exports.useEffect(()=>{e.bokehJson&&(o.current&&(o.current.innerHTML=""),window.Bokeh.embed.embed_item({doc:e.bokehJson,root_id:e.bokehJson.roots.root_ids[0],target_id:t}))},[e.bokehJson,t]),n.exports.createElement("div",{style:{padding:"10px",height:"100%",width:"100%"}},n.exports.createElement(d,{className:"bokeh-card",style:{width:"100%"}},e.headerElements,!!e.contentNotFound||!e.bokehJson?e.notFoundElements?e.notFoundElements:null:n.exports.createElement("div",{ref:o,id:t})))},m=e=>{const o=e.input,t=l(o);return t.loading?i.createElement("div",null):i.createElement(c,{bokehJson:t.contents?JSON.parse(t.contents):null})};export{m as default};
//# sourceMappingURL=Component.8131cbd1.js.map
