import{ah as i,u as o,W as e}from"./index.16929924.js";const c=l=>{const a=l.input,r=i({file:a}),t=o(r);return e.createElement("div",null,t.loading?e.createElement("div",null):e.createElement("img",{style:{maxWidth:"100%"},alt:"cool-alt",src:t.result}))};export{c as default};
//# sourceMappingURL=Component.294baeb1.js.map
